@@ -61,6 +61,7 @@ import           Basement.Compat.Primitive
 import           Basement.Monad
 import           Basement.PrimType
 import           Basement.Compat.Base
+import           Basement.Compat.C.Types
 import           Basement.Compat.Semigroup
 import qualified Basement.Runtime as Runtime
 import           Data.Proxy
@@ -74,7 +75,6 @@ import qualified Basement.Block as BLK
 import qualified Basement.Block.Mutable as MBLK
 import           Basement.Numerical.Additive
 import           Basement.Bindings.Memory
-import           Foreign.C.Types
 import           System.IO.Unsafe (unsafeDupablePerformIO)
 
 -- | A Mutable array of types built on top of GHC primitive.
@@ -90,6 +90,9 @@ data MUArrayBackend ty st = MUArrayMBA (MutableBlock ty st) | MUArrayAddr (Final
 instance PrimType ty => Alg.Indexable (Ptr ty) ty where
     index (Ptr addr) = primAddrIndex addr
 
+instance Alg.Indexable (Ptr Word8) Word64 where
+    index (Ptr addr) = primAddrIndex addr
+
 instance (PrimMonad prim, PrimType ty) => Alg.RandomAccess (Ptr ty) prim ty where
     read (Ptr addr) = primAddrRead addr
     write (Ptr addr) = primAddrWrite addr
@@ -171,6 +174,10 @@ isMutablePinned (MUArray _ _ (MUArrayMBA mb))  = BLK.isMutablePinned mb
 newPinned :: forall prim ty . (PrimMonad prim, PrimType ty) => CountOf ty -> prim (MUArray ty (PrimState prim))
 newPinned n = MUArray 0 n . MUArrayMBA <$> MBLK.newPinned n
 
+-- | Create a new unpinned mutable array of size @n elements.
+--
+-- If the size exceeds a GHC-defined threshold, then the memory will be
+-- pinned. To be certain about pinning status with small size, use 'newPinned'
 newUnpinned :: forall prim ty . (PrimMonad prim, PrimType ty) => CountOf ty -> prim (MUArray ty (PrimState prim))
 newUnpinned n = MUArray 0 n . MUArrayMBA <$> MBLK.new n
 
@@ -288,16 +295,15 @@ onBackendPure :: (Block ty -> a)
 onBackendPure goBA goAddr arr = onBackend goBA (\_ -> pureST . goAddr) arr
 {-# INLINE onBackendPure #-}
 
-onBackendPure' :: PrimType  ty
+onBackendPure' :: forall ty a . PrimType  ty
                => UArray ty
                -> (forall container. Alg.Indexable container ty 
                    => container -> Offset ty -> Offset ty -> a)
                -> a
-onBackendPure' arr f = onBackendPure (\c -> f c start end) 
-                                     (\c -> f c start end) arr
-  where !len = length arr
-        !start = offset arr
-        !end = start `offsetPlusE` len
+onBackendPure' arr f = onBackendPure f' f' arr
+  where f' :: Alg.Indexable container ty => container -> a
+        f' c = f c start end
+          where (ValidRange !start !end) = offsetsValidRange arr
 {-# INLINE onBackendPure' #-}
 
 onBackendPrim :: PrimMonad prim
@@ -319,12 +325,12 @@ onMutableBackend _     onAddr (MUArray _ _ (MUArrayAddr fptr)) = onAddr fptr
 {-# INLINE onMutableBackend #-}
 
 
-unsafeDewrap :: (ByteArray# -> Offset ty -> a)
+unsafeDewrap :: (Block ty -> Offset ty -> a)
              -> (Ptr ty -> Offset ty -> ST s a)
              -> UArray ty
              -> a
 unsafeDewrap _ g (UArray start _ (UArrayAddr fptr))     = withUnsafeFinalPtr fptr $ \ptr -> g ptr start
-unsafeDewrap f _ (UArray start _ (UArrayBA (Block ba))) = f ba start
+unsafeDewrap f _ (UArray start _ (UArrayBA ba)) = f ba start
 {-# INLINE unsafeDewrap #-}
 
 unsafeDewrap2 :: (ByteArray# -> ByteArray# -> a)
@@ -388,7 +394,7 @@ vToList a
     | otherwise = unsafeDewrap goBa goPtr a
   where
     !len = length a
-    goBa ba start = loop start
+    goBa (Block ba) start = loop start
       where
         !end = start `offsetPlusE` len
         loop !i | i == end  = []
@@ -601,29 +607,26 @@ append a b
     !la = length a
     !lb = length b
 
-concat :: PrimType ty => [UArray ty] -> UArray ty
-concat [] = empty
-concat l  =
-    case filterAndSum (CountOf 0) [] l of
-        (_,[])            -> empty
-        (_,[x])           -> x
-        (totalLen,chunks) -> runST $ do
-            r <- new totalLen
-            doCopy r (Offset 0) chunks
-            unsafeFreeze r
+concat :: forall ty . PrimType ty => [UArray ty] -> UArray ty
+concat original = runST $ do
+    r <- new total
+    goCopy r 0 original
+    unsafeFreeze r
   where
-    -- TODO would go faster not to reverse but pack from the end instead
-    filterAndSum !totalLen acc []     = (totalLen, List.reverse acc)
-    filterAndSum !totalLen acc (x:xs)
-        | len == CountOf 0 = filterAndSum totalLen acc xs
-        | otherwise      = filterAndSum (len+totalLen) (x:acc) xs
-      where len = length x
+    !total = size 0 original
+    -- size
+    size !sz []     = sz
+    size !sz (x:xs) = size (length x + sz) xs
 
-    doCopy _ _ []     = return ()
-    doCopy r i (x:xs) = do
-        unsafeCopyAtRO r i x (Offset 0) lx
-        doCopy r (i `offsetPlusE` lx) xs
-      where lx = length x
+    zero = Offset 0
+
+    goCopy r = loop
+      where
+        loop _  []      = pure ()
+        loop !i (x:xs) = do
+            unsafeCopyAtRO r i x zero lx
+            loop (i `offsetPlusE` lx) xs
+          where !lx = length x
 
 -- | Create a Block from a UArray.
 --
